# Copyright 2009 Canonical Ltd.
#
# This file is part of desktopcouch.
#
#  desktopcouch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# desktopcouch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with desktopcouch.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Rick Spencer <rick.spencer@canonical.com>

"""A Treeview for CouchDB"""

import gtk
import gobject
from server import CouchDatabase
from record import Record
from couchdb.client import Server

class CouchWidget(gtk.TreeView):
    __gtype_name__ = "CouchWidget"

    def __init__(self, server_ip = 'http://127.0.0.1:5984/'):
        """Create a new Couchwidget
        optional arguments:
        server_ip - specify and ip and port for the server location.
        Defaults to http://127.0.0.1:5984
        
        """

        gtk.TreeView.__init__(self)

        #assume a local server
        self.server = Server(server_ip)

        #set up the default values
        self.__db = None
        self.__record_type = None
        self.__headings = None
        self.__editable = False

    @property
    def headings(self):
        """ set headings - A list of strings to act as headings for the
        TreeView.

        Note that the CouchWidget uses headings for keys, and visa versa.
        If a record does not contain a value for a specified value
        the CouchWidget will simply display an empty string for the
        value. If the widget is set to editable, the user will be able
        to add values to the database.

        Setting this property will cause the widget to reload.        

        """

        return self.__headings

    @headings.setter
    def headings(self, headings):
        self.__headings = headings
        if self.record_type != None:
            self.set_record_type(self.record_type)

    @property
    def editable(self):
        """editable - bool value, True to make editable
        If set to True, changes are immediately
        persisted to the database.

        """
        return self.__editable

    @editable.setter
    def editable(self, editable):
        self.__editable = editable

        #refresh the treeview if possible
        if self.record_type != None:
            self.set_record_type(self.record_type)

    @property
    def database(self):
        """database - gets an instance to the CouchDB.
        Set to a string to change the database.

        """        
        return self.__db

    @database.setter
    def database(self, db_name):
        self.__db = CouchDatabase(db_name, create=True)

        if self.record_type != None:
            self.set_record_type(self.record_type)

    @property
    def record_type(self):
        """record_type - a string specifying the record type of
        the documents to retrieve from the CouchDB.

        Will cause the TreeView to refresh when set.
        """
        return self.__record_type

    @record_type.setter
    def record_type(self, record_type):

        #store the record type string
        self.__record_type = record_type

        #if the database is not set up, just return
        if self.__db == None:
            return

        #refresh the model structure
        self.__reset_model()

        #retrieve the docs for the record_type, if any
        filter = """function(doc) { if (doc.record_type == '%s') { emit(doc._id, doc); } }""" %record_type
        results = self.__db.query(filter)

        #if headings are already assigned, set up headings and columns
        if self.headings != None:
            self.__make_headings()

        first_row = True
        for r in results:
            #if headings are not set up, infer them from 
            #the first stored record
            if self.headings == None and first_row:
                self.headings = []
                for k in r.value.keys():
                    if not k.startswith("_") and k != "record_type":
                        self.headings.append(k)
                #now set up the columns and headers
                self.__make_headings()              
            first_row = False

            #lists have to match the list_store columns in length
            #so we have to make rows as long as the headerings
            #note that the last value is reserved for the doc_id
            col_count = len(self.headings) + 1
            row = ["" for i in xrange(col_count)]

            #replace the values in the row with the values in the
            #retrieved rows. If the the key is not in the doc, 
            #just continue with an empty string
            for i, h in enumerate(self.headings):
                try:
                    row[i] = r.value[h]
                except:
                    pass
    
            #set the last value as the document_id, and append
            row[-1] = r.key
            self.list_store.append(row)

        #apply the model tot he Treeview                 
        self.set_model(self.list_store)

    def append_row(self, new_row, auto_save = True):
        """append_row: add a row to the TreeView.

        arguments:
        new_row - a list of string values to add to the TreeView.
        Empty strings will appear as empty cells. It is not required
        to specify all values to include in the new row, as
        append_row will copy all the values starting at position 0
        and will fill in the remaining values with empty strings.
        
        auto_save - if True will immediately commit the row
        to the database. Otherwise, the row won't be saved until
        a cell is edited. Typically should be False if adding an
        empty row. An empty list will not be autosaved, even if 
        auto_save is set to True.

        Defaults to True.

        """

        #lists have to match the list_store columns in length
        #note that the last value is reserved for the doc_id
        col_count = len(self.headings) + 1
        row = ["" for i in xrange(col_count)]
        
        #create a document for autosaving
        doc = {"record_type":self.record_type}    
        
        #copy the values into the row
        #add to the document in case it needs to be auto saved
        for i, v in enumerate(new_row):
            row[i] = new_row[i]
            doc[self.headings[i]] = new_row[i]
        
        rec = Record(doc)

        #auto save non-empty rows
        if auto_save and len(new_row) > 0:
            doc_id = self.__db.put_record(rec)
            row[len(self.headings)] = doc_id #store the id
            
        #add it to the list store
        self.list_store.append(row)

    def __reset_model(self):
        """ __reset_model - internal funciton, do not call directly.
        This function is typically called when properties are set

        """

        #remove the current columns from the TreeView
        cols = self.get_columns()
        for c in cols:
            self.remove_column(c)

        #make a new ListStore
        col_count = len(self.headings) + 1
        self.list_store = gtk.ListStore(*[gobject.TYPE_STRING for i in xrange(col_count)])


    def __make_headings(self):
        """ __make_headings - internal funciton do not call direclty
        Sets up the headings and columnss for the TreeView.

        """

        for i, h in enumerate(self.headings):
            #set up cell renderer
            rend = gtk.CellRendererText()
            rend.mode = gtk.CELL_RENDERER_MODE_EDITABLE
            rend.set_property("editable", self.editable)
            #connect to save function
            rend.connect("edited", self.__edited,i)

            #create a column and add to the TreeView
            col = gtk.TreeViewColumn(h, rend, text=i)
            self.append_column(col)
        
    def __edited(self, cellrenderertext, path, new_text, col):
        """ __edited - internal signal handler.
        Updates the database if a cell in the Treeview
        has been edited.
        
        """

        #get an iterator that points to the edited row
        iter = self.list_store.get_iter(path)

        #update the ListStore with the new text
        self.list_store.set_value(iter, col, new_text)

        #get the key/heading for the edited cell
        #as well as the id
        key = self.headings[col]
        id = self.list_store.get_value(iter, len(self.headings))

        if id == "": #the row has not been stored
            #create a document
            doc = {"record_type":self.record_type}
           
            for i, h in enumerate(self.headings):
                #copy the all the values into the doc
                doc[h] = self.list_store.get_value(iter,i)      

            rec = Record(doc)
            #save the document and store the id
            doc_id = self.__db.put_record(rec)
            self.list_store.set_value(iter, len(self.headings), doc_id)

        else: #it has been saved
            self.__db.update_fields(id,{key:new_text})

if __name__ == "__main__":
    """creates a test CouchWidget if called directly"""

    #create and show a test window
    win = gtk.Window(gtk.WINDOW_TOPLEVEL)
    win.set_title("CouchWidget Test Window")
    win.connect("destroy",gtk.main_quit)
    win.show()

    #create a test widget with test database values
    cw = CouchWidget()
    cw.database = "couch_widget_test"

    #allow editing
    cw.editable = True

    #create headers/keys
    cw.headings = ["Key1","Key2","Key3","Key4"]

    #set the record_type for the TreeView
    #it will not populate without this value being set
    cw.record_type = "test_record_type"

    #create a row with all four columns set
    cw.append_row(["val1","val2","val3","val4"])

    #create a row with only the second column set
    cw.append_row(["","val2"])

    #create an empty row (which will not be saved until
    #the user edits it)
    cw.append_row([])

    #show the control, add it to the window, and run the main loop
    cw.show()
    win.add(cw)
    gtk.main()

