# Copyright 2009 Canonical Ltd.
#
# This file is part of desktopcouch.
#
#  desktopcouch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# desktopcouch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with desktopcouch.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Eric Casteleijn <eric.casteleijn@canonical.com>
#          Mark G. Saye <mark.saye@canonical.com>
#          Stuart Langridge <stuart.langridge@canonical.com>

"""The Desktop Couch Records API."""

from couchdb import Server
from couchdb.client import ResourceNotFound, ResourceConflict

from record import Record


class NoSuchDatabase(Exception):
    "Exception for trying to use a non-existent database"

    def __init__(self, dbname):
        self.database = dbname
        super(NoSuchDatabase, self).__init__()

    def __str__(self):
        return ("Database %s does not exist on this server. (Create it by "
                "passing create=True)") % self.database


class CouchDatabase(object):
    """An small records specific abstraction over a couch db database."""

    def __init__(self, database, uri=None, record_factory=None, create=False):
        if not uri:
            port = self._find_couchdb_port()
            self.server_uri = "http://localhost:%s" % port
        else:
            self.server_uri = uri
        self._server = Server(self.server_uri)
        if database not in self._server:
            if create:
                self._server.create(database)
            else:
                raise NoSuchDatabase(database)
        self.db = self._server[database]
        self.record_factory = record_factory or Record

    def _find_couchdb_port(self):
        """Look for a running Desktop CouchDB. If there isn't one, fall back
           to the system CouchDB on 5984."""
        # Currently using horrible lsof command; this will eventually use D-Bus
        # pylint: disable-msg=W0702
        try:
            import subprocess
            from desktopcouch import local_files
            lsofcmd = 'lsof -i4TCP@localhost | grep $(couchdb -C %s ' + \
               '-p %s -s | cut -d" " -f7 | cut -d "," -f1) 2>/dev/null | ' + \
                      'grep "TCP " | grep LISTEN | ' + \
                      'awk \'{print $8}\' | cut -d":" -f2'
            actual_lsof_cmd = lsofcmd % (local_files.FILE_INI,
              local_files.FILE_PID)
            stdout, stderr = subprocess.Popen(actual_lsof_cmd, shell=True,
              stdout=subprocess.PIPE).communicate()
            port = stdout.strip()
        except:
            port = 5984
        if not port:
            port = 5984
        return port

    def query(self, map_fun, reduce_fun=None, language='javascript',
      wrapper=None, **options):
        """Pass-through to CouchDB query"""
        return self.db.query(map_fun, reduce_fun, language,
              wrapper, **options)

    def get_record(self, record_id):
        """Get a record from back end storage."""
        try:
            couch_record = self.db[record_id]
        except ResourceNotFound:
            return None
        data = {}
        data.update(couch_record)
        record = self.record_factory(data=data)
        record.record_id = record_id
        return record

    def put_record(self, record):
        """Put a record in back end storage."""
        record_id = record.record_id or record._data.get('_id', '')
        record_data = record._data
        if record_id:
            self.db[record_id] = record_data
        else:
            record_id = self._add_record(record_data)
        return record_id

    def update_fields(self, doc_id, fields):
        """Safely update a number of fields. 'fields' being a
        dictionary with fieldname: value for only the fields we want
        to change the value of.
        """
        while True:
            doc = self.db[doc_id]
            doc.update(fields)
            try:
                self.db[doc.id] = doc
            except ResourceConflict:
                continue
            break

    def _add_record(self, data):
        """Add a new record to the storage backend."""
        return self.db.create(data)

    def delete_record(self, record_id):
        """Delete record with given id"""
        record = self.db[record_id]
        record.setdefault('application_annotations', {}).setdefault(
            'Ubuntu One', {}).setdefault('private_application_annotations', {})[
            'deleted'] = True
        self.db[record_id] = record

    def record_exists(self, record_id):
        """Check if record with given id exists"""
        if record_id not in self.db:
            return False
        record = self.db[record_id]
        return 'deleted' not in record.get('application_annotations', {}).get(
            'Ubuntu One', {}).get('private_application_annotations', {})


