# Copyright 2009 Canonical Ltd.
#
# This file is part of desktopcouch.
#
#  desktopcouch is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation.
#
# desktopcouch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with desktopcouch.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Eric Casteleijn <eric.casteleijn@canonical.com>

"""testing database/contact.py module"""

import testtools, utils

from desktopcouch.records.server import CouchDatabase
from desktopcouch.records.record import Record

FAKE_RECORD_TYPE = "http://example.org/test"

js = """
function(doc) {
    if (doc.record_type == '%s') {
        emit(doc._id, null);
    }
}""" % FAKE_RECORD_TYPE

top = utils.get_rootdir()
output = '%s/tmp/couch-test-output' % top


class TestCouchDatabase(testtools.TestCase):
    """tests specific for CouchDatabase"""

    def setUp(self):
        """setup each test"""
        # Connect to CouchDB server
        self.dbname = self._testMethodName
        self.database = CouchDatabase(self.dbname, create=True)

    def tearDown(self):
        """tear down each test"""
        del self.database._server[self.dbname]

    def test_get_record(self):
        """Test getting a record."""
        record = Record({'record_number': 0}, record_type="http://example.com/")
        record_id = self.database.put_record(record)
        retrieved_record = self.database.get_record(record_id)
        self.assertEqual(0, retrieved_record['record_number'])

    def test_put_record(self):
        """Test putting a record."""
        record = Record({'record_number': 0}, record_type="http://example.com/")
        record_id = self.database.put_record(record)
        retrieved_record = self.database._server[self.dbname][record_id]
        self.assertEqual(
            record['record_number'], retrieved_record['record_number'])

    def test_delete_record(self):
        """Test deletion of records."""
        record = Record({'record_number': 0}, record_type="http://example.com/")
        record_id = self.database.put_record(record)
        self.database.delete_record(record_id)
        deleted_record = self.database._server[self.dbname][record_id]
        self.assert_(deleted_record['application_annotations']['Ubuntu One'][
            'private_application_annotations']['deleted'])

    def test_record_exists(self):
        """Test checking whether a record exists."""
        record = Record({'record_number': 0}, record_type="http://example.com/")
        record_id = self.database.put_record(record)
        self.assert_(self.database.record_exists(record_id))
        self.database.delete_record(record_id)
        self.assert_(not self.database.record_exists(record_id))

    def test_update_fields(self):
        """Test the update_fields method."""
        dictionary = {'record_number': 0, 'field1': 1, 'field2': 2}
        record = Record(dictionary, record_type="http://example.com/")
        record_id = self.database.put_record(record)
        # manipulate the database 'out of view'
        non_working_copy = self.database.get_record(record_id)
        non_working_copy['field2'] = 22
        non_working_copy['field3'] = 3
        self.database.put_record(non_working_copy)
        self.database.update_fields(record_id, {'field1': 11})
        working_copy = self.database.get_record(record_id)
        self.assertEqual(0, working_copy['record_number'])
        self.assertEqual(11, working_copy['field1'])
        self.assertEqual(22, working_copy['field2'])
        self.assertEqual(3, working_copy['field3'])
